import java.util.*;
import java.util.logging.*;

public class ASTAffine extends ASTNode {
    String ch;
    ASTNode rhs;

    HashSet<String> usageSet;
    HashSet<String> coaffineSet;

    public ASTAffine (String _ch,ASTNode _rhs) {
        ch = _ch;
        rhs = _rhs;
        usageSet = new HashSet<String>();
        coaffineSet = new HashSet<String>();
    }

    public  void ASTInsertUse(String _ch, ASTNode here, Boolean disCont) throws Exception
    {
     if(disCont && !_ch.equals(ch))
         anc.ASTInsertUse(_ch,this,disCont);
     else {
         ASTNode pushUse = new ASTUse(_ch, here);
         pushUse.eg = eg;
         here.setanc(pushUse);
         pushUse.setanc(this);
         rhs = pushUse;
     }
    }

    public void ASTInsertCall(String ch, String cho, ASTType t, ASTNode here) throws Exception{
        ASTNode pushCall = new ASTCall(ch, cho, t,here);
        pushCall.eg = eg;
        here.setanc(pushCall);
        pushCall.setanc(this);
        rhs = pushCall;
    }

    public void ASTInsertWhyNot(String _ch, ASTType _t, ASTNode here) throws Exception
    {
        if(_ch.equals(ch)) {
            ASTNode pushWhy = new ASTWhy(_ch, here);
            pushWhy.eg = eg; 
            here.setanc(pushWhy);
            pushWhy.setanc(this);
            rhs = pushWhy;
            eg.insert(_ch,_t);
        } else
            anc.ASTInsertWhyNot(_ch, _t, this);
    }
    public void ASTupdCont(ASTNode newCont, ASTNode caller) throws Exception{
        rhs = newCont;
    }

    public ASTNode ASTweakeningOnLeaf(String _ch, boolean exp) throws Exception
    {
	if (exp)
	    return this.ASTweakeningHere(_ch,  exp); // must insert here because affine blocks linear names ?x
	else {
	    rhs = rhs.ASTweakeningOnLeaf(_ch,exp);
	    // if(coaffineSet.contains(_ch)) System.out.println("in CoaffineSet");
	    coaffineSet.add(_ch);
	    return this;
	}
    }

    public void typecheck(Env<ASTType> ed, Env<ASTType> eg, Env<EnvEntry> ep) throws Exception {
        this.eg = eg;
        this.inferUses(ch,ed,ep);
        ASTType ty = ed.find(ch);
        ty = ty.unfoldType(ep);
 	    ty = ASTType.unfoldRec(ty);
        if (ty instanceof ASTAffineT) {
            Set<String> s = rhs.fn(new HashSet<String>());
            s.remove(ch);
            Iterator<String> it = s.iterator();
            while(it.hasNext()){
                String id = it.next();
                ASTType tyId = null;
                try {
                    tyId = ed.find(id);
                } catch(Exception e) {}
                if (tyId != null) {
                    tyId = tyId.unfoldType(ep);
                    tyId = ASTType.unfoldRec(tyId);
		    // System.out.println(tyId.toStr(ep));
                    if (tyId instanceof ASTUsageT)
                        usageSet.add(id);
                    else if (tyId instanceof ASTCoAffineT)
                        coaffineSet.add(id);
                    else if (tyId instanceof ASTWhyT){
                        //System.out.println("DEBUG !: " + id + " is of type ?");
                        ASTWhyT t = (ASTWhyT) tyId;
                        tyId = t.getin();
                        this.getanc().ASTInsertWhyNot(id,tyId, this);
                        ed.updmove(id);
                        //System.out.println("DEBUG ! infer ? for " + id);
                        //eg.crawl();
                    }
                    else
                        throw new TypeError("AFFINE " + ch + ": " + id + " is neither of USAGE nor of COAFFINE type.");
                }
            };


            ASTAffineT tyr = (ASTAffineT)ty;
            ed.upd(ch, tyr.getin().unfoldType(ep));
            rhs.typecheck(ed,eg,ep);
            rhs.linclose(ed,ep);
        }
        else throw new TypeError("AFFINE: "+ch+" is not of AFFINE type: found "+ty.toStr(ep));
    }

    public Set<String> fn(Set<String> s) {
        s = rhs.fn(s);
        s.add(ch);
        return s;
    }

    public Set<String> fnLinear(Set<String> s) {
        s = rhs.fnLinear(s);
        s.add(ch);
        return s;
    }

            public ASTNode subst(Env<ASTType> e) {
	    ASTAffine p = new ASTAffine(ch,rhs.subst(e));
	    p.rhs.setanc(p);
	    return p;
	}

    public void subs(String x, String y){// implements x/y (substitutes y by x)
        if(y == ch)
            ch = x;

        rhs.subs(x,y);
    }
    public void runproc(Env<EnvEntry> ep, Env<LinSession> ed, Env<Server> eg, Logger logger) throws Exception{

        Channel channel = (Channel) ed.find(ch);
        String selection = (String) channel.receive();
        if(selection == "USE"){
            rhs.runproc(ep,ed,eg,logger);
        }else if (selection == "DISCARD"){

            Iterator<String> itU = usageSet.iterator();
            while(itU.hasNext()){
                String id = itU.next();
                CLLSj.threadPool.submit( new Runnable(){
                    public void run(){ try {
                        (new ASTRelease(id)).runproc(ep, ed, eg, logger);
                    } catch (Exception e) {e.printStackTrace(System.out);} }
                });
            }

            Iterator<String> itC = coaffineSet.iterator();
            while(itC.hasNext()){
                String id = itC.next();
                CLLSj.threadPool.submit( new Runnable(){
                    public void run(){ try {
                        (new ASTDiscard(id)).runproc(ep, ed, eg, logger);
                    } catch (Exception e) {e.printStackTrace(System.out);} }
                });
            }

        }


    }

    public String toStr(Env<EnvEntry> ep) throws Exception {
        return "affine " + ch + ";\n" + rhs.toStr(ep);
    }

    	public void show()  {
	    System.out.println(this+" "+anc);
	    rhs.show();
	}
    
}
