PARSER_BEGIN(CLLSj)

import java.util.*;
import java.io.*;
import java.util.logging.*;
import java.util.concurrent.*;


public class CLLSj {

static CLLSj parser;
static Env<ASTType> ed;
static Env<ASTType> eg;
static Env<EnvEntry> ep;

static boolean unsafe = true;

//public static ConsoleHandler handler = new ConsoleHandler() {{setOutputStream(System.out);}};

//public static Logger logger = Logger.getLogger("ASTNode");

public static final ExecutorService threadPool = Executors.newCachedThreadPool();
//public static final ExecutorService threadPool = Executors.newFixedThreadPool(10000);

static {
    try {
    System.setProperty("java.util.logging.SimpleFormatter.format", "%5$s %n");
    }
    catch (Exception e) { e.printStackTrace(System.out); }
}

public static void main(String args[]) {
    ed = new Env<ASTType>();
    eg = new Env<ASTType>();
    ep = new Env<EnvEntry>();
    FileInputStream fboot;

    //System.out.println("maximum pool size = " + ((ThreadPoolExecutor) threadPool).getMaximumPoolSize() );
    try {
     fboot = new FileInputStream("lib/init.clls");
    } catch (FileNotFoundException ex) {

System.out.println("Cannot find lib/init.cls");
      return;
    };
    processFile(System.getProperty("user.dir"),new CLLSj(fboot), true);
    System.out.println ("CLASS version 7.0, 22 November 2022\n");
    unsafe = false;
    processFile(System.getProperty("user.dir"),new CLLSj(System.in), false);
    System.exit(0);
}

public static void processFile(String path, CLLSj parser, boolean inc)
{
    while (true) {
    if (!inc) { System.out.print("> "); }
    try {
      //ep.crawl();
	  ASTNode n = parser.Start();
   	  eg = new Env<ASTType>(); // because of in plave inserts durimg inference
   	  if (n == null) {  return; }
   	  if (n instanceof ASTInclude) {
	      String fd = ((ASTInclude)n).getFn();
	      //System.out.println("FD="+fd);
	      //System.out.println("PATH="+path);
	      String fn = path+"/"+fd;
	      String newpath = fn.substring(0,fn.lastIndexOf("/"));
	      //System.out.println("NEWPATH="+newpath);
	      InputStream f = new FileInputStream(fn);
	      System.out.println("> including "+fn);
	      processFile(newpath, new CLLSj(f), true); // this should return on error
	      f.close();
	      if (!unsafe)
	      System.out.println("included "+fn);
    	  }
    	  else {
       //  n.typecheck(ed,eg,ep);
    	   if (n instanceof ASTDList) {
       	   ASTDList ld = (ASTDList)n;
	   ld.typecheckmany(ed,eg,ep);
	   ep = ld.definemany(ed,eg,ep);
           //  n.typecheck(ed,eg,ep); // recursion !
           //  ep=ep.assoc(td.id,new TypeDefEntry(td));
      	   //  if (!unsafe) System.out.println ("Type "+td.id+": defined.");
       	   } else
    	   if (n instanceof ASTTypeDef) {
       	   ASTTypeDef td = (ASTTypeDef)n;
           n.typecheck(ed,eg,ep); // recursion !
           ep=ep.assoc(td.id,new TypeDefEntry(td));
      	   if (!unsafe)
	      System.out.println ("Type "+td.id+": defined.");
       	   } else
	   if (n instanceof ASTPList) {
       	      ASTPList ld = (ASTPList)n;
	      ld.typecheckmany(ed,eg,ep);
	      ep = ld.definemany(ed,eg,ep,unsafe);
	   }
	   else
       	   if (n instanceof ASTProcDef) {
        ASTProcDef td = (ASTProcDef)n;
        n.typecheck(ed,eg,ep);
	   	 ep=ep.assoc(td.id,new ProcEntry(td));
           	 td.setEnv(ep);
	      	 if (!unsafe)
		        System.out.println ("Process  "+td.id+": defined.");
	      }
	      else
	      { // must be process to run

                /* some code (e.g. argument-as-expressions) assume that every node has a parent,
                so we will create a dummy coclose node to serve as parent of the running process n */
                ASTCoClose parent = new ASTCoClose("dummy", n);
                n.setanc(parent);

              	n.typecheck(ed,eg,ep);
		// System.out.println("Type Checked.");
		
            	if (n.repeat > 1) System.out.println("[RunStatus] Running process " + n.repeat + " times.");

             Logger logger = Logger.getAnonymousLogger();
             ConsoleHandler handler = new ConsoleHandler() {{setOutputStream(System.out);}};
             handler.setFormatter(new SimpleFormatter());
             logger.addHandler(handler);

             if(n.info){
                logger.setLevel(Level.INFO);
             }
             else
                logger.setLevel(Level.OFF);

             int maxThreads = 0;
             while(n.repeat > 0){
	  	        parent.rhs.runproc(ep, new Env<LinSession>(), new Env<Server>(), logger);
			    Thread.sleep(1);
			    while (((ThreadPoolExecutor)threadPool).getActiveCount()>0) {
                if(((ThreadPoolExecutor)threadPool).getActiveCount() > maxThreads){
                    maxThreads = ((ThreadPoolExecutor)threadPool).getActiveCount();
                    //System.out.println("maxThreads = " + maxThreads);
}
                 //System.out.println("number of active threads= " + ((ThreadPoolExecutor) threadPool).getActiveCount() );
                //System.out.println("pool size= " + ((ThreadPoolExecutor) threadPool).getPoolSize() );

                Thread.sleep(1);};
                //System.out.println("getActiveCount = 0");
                //System.out.println("maxThreads = " + maxThreads);
                n.repeat--;
             }
             //System.out.println("Putting thread to sleep");
             //Thread.sleep(4000);
             //System.out.println("Finished Puttin thread to sleep");
		     //System.out.println("[RunStatus] Finished running process.");
		 }
		 }
	} catch (TypeError e) {
    	  	System.out.println ("TypeError: ("+e.msg+")");
	  	e.printStackTrace(System.out);	
		parser.ReInit(System.in);
		}
	catch (KindError e) {
    	        System.out.println ("KindError: "+e.msg);
		if (inc) { return; } else {
	      	parser.ReInit(System.in);
		}
//	} catch (SyntaxError e) {
//    	  	parser.ReInit(System.in);
//	  	System.out.println (e.msg);
	} catch (FileNotFoundException e) {
	  System.out.println ("File not found: "+e.getMessage());
		if (inc) { return; } else {
	      	parser.ReInit(System.in);
		}
	} catch (TokenMgrError e) {
	  System.out.println(e.getMessage());
		if (inc) { return; } else {
	      	parser.ReInit(System.in);
		}
	} catch (ParseException e) {
 	  System.out.println(e.getMessage());
		if (inc) { return; } else {
	      	parser.ReInit(System.in);
		}
	} catch (Exception e) {
	  System.out.println ("Unexpected error: ");
	  e.printStackTrace(System.out);
		if (inc) { return; } else {
	      	parser.ReInit(System.in);
		}
	  }
      }
    }
}


PARSER_END(CLLSj)

SKIP :
{
  " "
| "\t"
| "\r"
| "\n"
| < "//" (~["\r", "\n"])* >
| < "/*" > : ML_COM_NEXT
}

<ML_COM_NEXT> SKIP :
{
  < "*/" > : DEFAULT
| < ~[] >   
}

TOKEN :
{
  < FOR: "for" >
  |
  < TO: "to" >
  |
  < LET: "let" >
  |
  < LETC: "letc" >
  |
  < IN: "in" >
  |
  < REC: "rec" >
  |
  < UNSAFE_REC: "unsafe_rec" >
  |
  < COREC: "corec" >
  |
  < LETB: "let!" >
  |
  < INT: "int" >
  |
  < COINT: "coint" >
  |
  < LINT: "lint" >
  |
  < LCOINT: "colint" >
  |
   < LBOOL: "lbool" >
  |
  < LCOBOOL: "colbool" >
  |
  < LSTRING: "lstring" >
  |
  < LCOSTRING: "colstring" >
  |
  < IF: "if" >
  |
  < PRINT: "print" >
  |
  < PRINTLN: "println" >
  |
  < INFO: "trace" >
  |
  < REPEAT: "repeat" >
  |
  < SEND: "send" >
  |
  < INCLUDE: "include" >
  |
  < RECV: "recv" >
  |
  < SENDT: "sendty" >
  |
  < RECVT: "recvty" >
  |
  < DISCARD: "discard" >
  |
  < USE: "use" >
  |
  < SOME: "some" >
  |
  < RELEASE: "release" >
  |
  < TAKE: "take" >
  |
  < PUT: "put" >
  |
  < FWD: "fwd" >
  |
  < FWDB: "fwd!" >
  |
  < CUT: "cut" >
  |
  < MIX: "par" >
  |
  < CASE: "case" >
  |
  < CHOICE: "choice" >
  |
  < OFFER: "offer" >
  |  
  < OF: "of" >
  |
  < TYPE: "type" >
  |
  < PROC: "proc" >
  |
  < STATE: "state" >
  |
  < STATEL: "statel" >
  |
  < USAGE: "usage" >
  |
  < USAGEL: "usagel" >
  |
  < STATEB: "state!" >
  |
  < USAGEB: "usage!" >
  |
  <STATEBL: "state!l">
  |
  < USAGEBL: "usage!l">
  |
  < AFFINE: "affine" >
  |
  < COAFFINE: "coaffine" >
  |
  < CELL: "cell" >
  |
  < RD: "rd" >
  |
  < WRT: "wrt" >
  |
  < LOCK: "lock" >
  |
  < UNLK: "unlk" >
  |
  < CELLB: "cell!" >
  |
  < EMPTY: "empty" >
  |
  < CLOSE: "close" >
  |
  < CCLOSE: "wait" >
  |
  < SHARE: "share" >
  |
  < SHAREL: "shareL" >
  |
  < SHARER: "shareR" >
  |
  < UNFOLD: "unfold" >
  |
  < QUIT: "quit" >
  |
  < SLEEP: "sleep" >
  |
  < CALL: "call" >
  |
  < OR : "or" >
  |
  < AND : "and" >
  |
  < REPL: "bang" >
  |
  < TRUE: "true" >
  |
  < FALSE: "false" >
  |
  < RAND: "rand" >
  |
  < lab: "#"( ["A"-"Z", "a"-"z","_","0"-"9"] )+ >
  |
  < Id: ["A"-"Z", "a"-"z"] ( ["A"-"Z", "a"-"z","_", "0"-"9"] )* >
  |
  < Num: (["0"-"9"]) + >
  |
  < VBAR : "|" >
  |
  < PLUS : "+" >
  |
  < MINUS : "-">
  |
  < TIMES : "*">
  |
  < DIV : "/">
  |
  < NOT : "~">
  |
  < DOT : ".">
  |
  < BANG : "!">
  |
  < WHY : "?">
  |
  < WHYY : "??">
  |
  <LPAR : "(" >
  |
  <RPAR : ")" >
  |
  <LBRA : "{" >
  |
  <RBRA : "}" >
  |
  <LE : "<=" >
  |
  <GE : ">=" >
  |
  <LT : "<" >
  |
  <GT : ">" >
  |
  <EQ : "==" >
  |
  <NEQ : "!=" >
  |
  <SEMIC : ";" >
  |
  <COMMA : "," >
  |
  <COL : ":" >
  |
  <TERM: ";;" >
  |
  <STRING: "\"" (~["\""])* "\"" >
}

ASTNode Start() throws Exception:
{ASTNode n;
Token c;
List<ASTTypeDef> ld;
}
{
(   n=TypeD() <TERM>
  | n=ProcD() <TERM>
  | n=Proc() <TERM>
  | <REPEAT> c = <Num> n = Proc() <TERM> { n.repeat = Integer.parseInt(c.image); }
  | <INFO> n = Proc() <TERM> { n.info = true; }
  | <QUIT> <TERM> {n = null;}
  | <EOF> {n = null;}
  | <INCLUDE> c=<STRING> <TERM> {n = new ASTInclude(c.image);}
)
{return n;}
}

ASTTypeDef Type1D(boolean rec, boolean corec) throws Exception:
{ASTType p1; Token c1;
ASTTypeDef p;}
{
     c1 = <Id>
{ p = new ASTTypeDef(c1.image); }
(<LPAR> ParamListt(p) <RPAR>)? <LBRA>
         p1 = Type() { p.setType(p1); }
     <RBRA>
     { p.setRec(rec); p.setCoRec(corec);}
     {return p;}
}


ASTNode TypeD() throws Exception:
{ASTType p1; Token c1;
boolean rec = false;
boolean corec = false;
List<ASTTypeDef> ld = new ArrayList<ASTTypeDef>();
ASTTypeDef  p;}
{
     <TYPE>
     ( <REC> { rec = true; } | <COREC> { corec = true; })?
     p = Type1D(rec,corec) { ld.add(p); }
     (<AND> p = Type1D(rec,corec) { ld.add(p); })*
     {return new ASTDList(ld, rec || corec);}
}

void ParamListt(ASTTypeDef p) throws Exception:
{ASTType t; Token i;}
{
    ( i=<Id> {  p.addPar(i.image);  }
   (<COMMA> i=<Id> {  p.addPar(i.image);  } )*)*
}

void ArgListT(ASTIdT t) throws Exception:
{ASTType ty;}
{
    ( ty=Type() {t.addarg(ty); } (<COMMA> ty=Type() {t.addarg(ty); })*)*

}

ASTType Type() throws Exception:
{ ASTType l,r,t; Token s;}
{
(   s=<Id>  { t = new ASTIdT(s.image); } ( <LPAR> ArgListT((ASTIdT)t) <RPAR> )?
    |
    <LINT> {
    t = new ASTLintT();
    }
    |
    <LCOINT> {
    t = new ASTLCointT();
    }
    |
    <LBOOL> {
    t = new ASTLboolT();
    }
    |
    <LCOBOOL> {
    t = new ASTCoLboolT();
    }
    |
    <LSTRING> {
    t = new ASTLstringT();
    }
    |
    <LCOSTRING> {
    t = new ASTCoLstringT();
    }
    |
    <SEND> l=Type() <SEMIC> r=Type() { t = new ASTSendT(l,r); }
    |
    <RECV> l=Type() <SEMIC> r=Type() { t = new ASTRecvT(l,r); }
    |
    <SENDT> s=<Id> <SEMIC> r=Type() { t = new ASTSendTT(s.image,r); }
    |
    <RECVT> s=<Id> <SEMIC> r=Type() { t = new ASTRecvTT(s.image,r); }
    |
    <CHOICE> <OF> <LBRA>
    { ASTCaseT tc = new ASTCaseT();
    } (TypeCase(tc))+ <RBRA> { t = tc; }
    |
    <OFFER> <OF> <LBRA>
    { ASTOfferT to = new ASTOfferT();
    } (TypeOffer(to))+ <RBRA> { t = to; }
    |
    <CLOSE> { t = new ASTOneT(); }
    |
    <CCLOSE> { t = new ASTBotT(); }
    |
    <USAGE> l=Type()  { t = new ASTUsageT(l); }
    |
    <USAGEL> l=Type()  { t = new ASTUsageLT(l); }
    |
    <STATE> l=Type()  { t = new ASTCellT(l); }
    |
    <STATEL> l=Type()  { t = new ASTCellLT(l); }
    |
    <USAGEB> l=Type()  { t = new ASTUsageBT(l); }
    |
    <USAGEBL> l=Type()  { t = new ASTUsageBLT(l); }
    |
    <STATEB> l=Type()  { t = new ASTCellBT(l); }
    |
    <STATEBL> l=Type()  { t = new ASTCellBLT(l); }
    |
    <AFFINE> l=Type() { t = new ASTAffineT(l); }
    |
    <COAFFINE> l=Type() { t = new ASTCoAffineT(l); }
    |
    <NOT> l=Type() { t = new ASTNotT(l); }
    |
    <BANG> l=Type() { t = new ASTBangT(l); }
    |
    <WHY> l=Type() { t = new ASTWhyT(l); }
    |
    <LBRA> t=Type() <RBRA>
)
{ return t;}
}

void TypeCase(ASTCaseT lt) throws Exception:
{Token s; ASTType t;}
{
<VBAR> s=<lab> <COL> t=Type() {
   lt.addCase(s.image,t);
}
}

void TypeOffer(ASTOfferT lt) throws Exception:
{Token s; ASTType t;}
{
<VBAR> s=<lab> <COL> t=Type() {
   lt.addCase(s.image,t);
}
}

ASTProcDef Proc1D(boolean rec) throws Exception:
{ASTNode p1; Token c1;
ASTProcDef p = new ASTProcDef();}
{
     c1 = <Id>
     { p.setId(c1.image); }
     (<LT> TParamList(p) <GT>)?
     <LPAR> ParamList(p) <RPAR> <LBRA>
         p1 = Proc()
	 { p.setProc(p1);
	   p1.setanc(p);
	   p.setRec(rec); }
     <RBRA>
     {return p;}
}


ASTNode ProcD() throws Exception:
{ASTNode p1; Token c1;
boolean rec = false;
boolean unsafe_rec = false;
List<ASTProcDef> ld = new ArrayList<ASTProcDef>();
ASTProcDef p ;
}
{
     <PROC>
        ( <REC> {rec = true; } | <UNSAFE_REC> { unsafe_rec = true;})?
	 p = Proc1D(rec) { ld.add(p);  p.setUnsafe_Rec(unsafe_rec);  }
     (<AND> p = Proc1D(rec) { ld.add(p); })*
     {return new ASTPList(ld, rec, unsafe_rec );}
}

void ParamList(ASTProcDef p) throws Exception:
{ASTType t; Token i;}
{
   (
      i=<Id> <COL> t=Type() {  p.addDPar(i.image,t);  }
     (<COMMA> i=<Id> <COL> t=Type() {  p.addDPar(i.image,t);  } ) *
   )?
   ((<SEMIC>
     i=<Id> <COL> t=Type() {  p.addGPar(i.image,t);  }
    (<COMMA> i=<Id> <COL> t=Type() {  p.addGPar(i.image,t);  } ) *
   )?)
}

void TParamList(ASTProcDef p) throws Exception:
{ASTType t; Token i;}
{
    ( i=<Id> {  p.addTPar(i.image);  }
   (<COMMA> i=<Id> {  p.addTPar(i.image);  } )*)?
}

void ArgList(ASTId t) throws Exception :
{ASTExpr pe;}
{
    ( pe=Expr() {t.addExpr(pe); } (<COMMA> pe=Expr() {t.addExpr(pe); })*)?
    ( <SEMIC>  ( pe=Expr() {t.addGExpr(pe); } (<COMMA> pe=Expr() {t.addGExpr(pe); })*))?
}

void ArgTList(ASTId t) throws Exception:
{ASTType ty;}
{
    ( ty=Type() {t.addTpar(ty); } (<COMMA> ty=Type() {t.addTpar(ty); })*)?
}

ASTExpr Expr()
throws Exception:
{
ASTExpr p1,p2,p;
ASTType t;
Token c1,c2;
}
{
	p1=CExpr()
	(
	<AND> p2=CExpr() { p1 = new ASTAnd(p1,p2); }
	|
	<OR> p2=CExpr() { p1 = new ASTOr(p1,p2); }
	)?
	{ return p1; }
}


ASTExpr CExpr()
throws Exception:
{
ASTExpr p1,p2,p;
ASTType t;
Token c1,c2;
}
{
	p1=AExpr()
	(
	<EQ> p2=AExpr() { p = new ASTEq(p1,p2); p1.setanc(p); p2.setanc(p); p1 = p;  }
	|
	<NEQ> p2=AExpr() { p = new ASTNEq(p1,p2); p1.setanc(p); p2.setanc(p); p1 = p; }
	|
	<LT> p2=AExpr() { p = new ASTLt(p1,p2); p1.setanc(p); p2.setanc(p); p1 = p; }
	|
	<GT> p2=AExpr() { p = new ASTGt(p1,p2); p1.setanc(p); p2.setanc(p); p1 = p; }
	)?
	{ return p1; }
}


ASTExpr AExpr()
throws Exception:
{
ASTExpr p1,p2,p;
ASTType t;
Token c1,c2;
}
{
	p1=Term()
	(
	<PLUS> p2=Term() { p = new ASTAdd(p1,p2); p1.setanc(p); p2.setanc(p); p1 = p;}
	|
	<MINUS> p2=Term() { p=new ASTSub(p1,p2); p1.setanc(p); p2.setanc(p); p1 = p;}
	)* 
	{ return p1; }
}

ASTExpr Term()
throws Exception:
{
ASTExpr p1,p2,p;
ASTType t;
Token c1,c2;
}
{
	p1=Fact()
	(
	<TIMES> p2=Fact() { p = new ASTMul(p1,p2); p1.setanc(p); p2.setanc(p); p1 = p;}
	|
	<DIV> p2=Fact() { p=new ASTDiv(p1,p2); p1.setanc(p); p2.setanc(p); p1 = p;}
	)* 
	{ return p1; }
}

ASTExpr Fact()
throws Exception:
{
ASTExpr p1,p2,p;
ASTType t;
Token c1,c2;
}
{
(
	c1=<Num> { p = new ASTInt(Integer.parseInt(c1.image)); }
	|
	c1=<TRUE> { p = new ASTBool(true); }
	|
	c1=<FALSE> { p = new ASTBool(false); }
	|
	c1=<STRING> { p = new ASTString(c1.image.substring(1,c1.image.length()-1)); }
	|
	c1=<Id> { p = new ASTVId(c1.image); }
	|
	<LPAR> p = Expr() <RPAR>
	|
	<MINUS> p = Fact() { p1 = new ASTSub(new ASTInt(0),p); p.setanc(p1); p = p1; }
	|
	<BANG> p = Fact() { p1 = new ASTNot(p); p.setanc(p1); p = p1; }
)
	{ return p; }
}


ASTNode CCut() throws Exception:
{
ASTNode p1,p2,p;
ASTExpr pe;
ASTType t;
Token c1,c2;
String str;
}
{
 p1=Proc() (<VBAR> c1=<Id><COL> t=Type() <VBAR> p2=CCut()
 { p = new ASTCut(c1.image,t,p1,p2);
   p1.setanc(p);
   p2.setanc(p);
   p1 = p; })?
 { return p1; }

}

ASTNode Proc() throws Exception:
{
ASTNode p1,p2,p;
ASTExpr pe;
ASTType t = null;
Token c1,c2,c3,c4,c5;
String str;
}
{
(
    <PRINT>
    pe=Expr()
    <SEMIC>
    p2=Proc()
    { 
      p = new ASTPrintLn(pe,p2,false);
      p2.setanc(p);
      pe.setanc(p);
    }
    |
    <SLEEP>
    c1 = <Num>
    <SEMIC>
    p=Proc()
    {
      p1 = new ASTSleep(Integer.parseInt(c1.image),p);
      p.setanc(p1);
      p = p1;
    }
    |
    <PRINTLN>
    pe=Expr()
    <SEMIC>
    p2=Proc()
    { 
      p = new ASTPrintLn(pe,p2,true);
      pe.setanc(p);
      p2.setanc(p);
    }
    |
    <IF>
    pe=Expr()
    <LBRA>
    p1=Proc()
    <RBRA>
    <LBRA>
    p2=Proc()
    <RBRA>
    { 
      p = new ASTIf(pe,p1,p2);
      pe.setanc(p);
      p1.setanc(p);
      p2.setanc(p);
    }
    |
    <LET>
    c2=<Id>
    pe = Expr()  
    { 
    p = new ASTCoExpr(c2.image, pe);
    pe.setanc(p);
    }
    |
    <LETB>
    c2=<Id>
    pe = Expr()  
    { 
    p = new ASTPromoCoExpr(c2.image, pe);
    pe.setanc(p);
    }
    |
    <LPAR> <RPAR>
    { p = new ASTEmpty();}
    |
    <LBRA> p=Proc() <RBRA>
    |
    c1=<Id>  { p = new ASTId(c1.image);}
    (<LT>  ArgTList((ASTId)p) <GT>)?
    <LPAR> ArgList((ASTId)p) <RPAR>    
    |
    <CUT> <LBRA> p1=CCut() <RBRA>
    { p = p1;}
    |
    <LETC> <VBAR> c1=<Id><COL> t=Type() <VBAR> p1=Proc()
    <IN> p2=Proc()
    	  {   p = new ASTCut(c1.image,t,p2,p1);
  	      p1.setanc(p);
	      p2.setanc(p);
  	  }
    |
    <MIX> <LBRA> p1=Proc() ( <VBAR><VBAR> p2=Proc()
    	  { p = new ASTMix(p1,p2); p1.setanc(p); p2.setanc(p); p1=p; } )+ <RBRA>
    { p = p1; } // remove ??
    |
    <SHARE> c1=<Id> <LBRA> p1=Proc() ( <VBAR><VBAR> p2=Proc()
    	    { p =new ASTShare(c1.image,p1,p2); p1.setanc(p); p2.setanc(p); p1=p; } )+ <RBRA>
    { p = p1;} // remove ??
    |    
//    <SHARE> c1=<Id> <LBRA> p1=Proc() <VBAR> <VBAR> p2=Proc() <RBRA>
//    { p = new ASTShare(c1.image,p1,p2);}
//    |
    <SHAREL> c1=<Id> <LBRA> p1=Proc() <VBAR> <VBAR> p2=Proc() <RBRA>
    { p = new ASTShareL(c1.image,p1,p2); p1.setanc(p); p2.setanc(p); }
    |
    <SHARER> c1=<Id> <LBRA> p1=Proc() <VBAR> <VBAR> p2=Proc() <RBRA>
    { p = new ASTShareR(c1.image,p1,p2); p1.setanc(p); p2.setanc(p); }
    |
    <SEND> c1=<Id> <LPAR>
    (LOOKAHEAD(2)
     c2=<Id> (<COL> t=Type())? <DOT> p1=Proc() <RPAR> <SEMIC> p2=Proc()
    {
	p = new ASTSend(c1.image,c2.image,t,p1,p2);
	p1.setanc(p); p2.setanc(p); }
    |
    pe = Expr()  <RPAR> <SEMIC> p2=Proc()
    {
       p = new ASTSend(c1.image,ASTType.gensym(),null,pe,p2);
       pe.setanc(p); p2.setanc(p); 
    }
    )    
    |
    <RECV> c1=<Id> <LPAR> c2=<Id> ( <COL> t=Type()) ? <RPAR> <SEMIC> p1=Proc()
    { p = new ASTRecv(c1.image,c2.image,t,p1);	p1.setanc(p); }
    |
    <SENDT> c1=<Id> <LPAR> t=Type() <RPAR> <SEMIC> p2=Proc()
    { p = new ASTSendTy(c1.image,t,p2) ; p2.setanc(p);}
    |
    <RECVT> c1=<Id> <LPAR> c2=<Id> <RPAR> <SEMIC> p1=Proc()
    { p = new ASTRecvTy(c1.image,c2.image,p1); p1.setanc(p);}
    |
    <CASE> c1=<Id> <OF> <LBRA>
    { ASTCase tc = new ASTCase(c1.image);
    }  (ProcCase(tc))+ <RBRA> { p = tc; }
    |
    c1=<lab> c2=<Id> <SEMIC> p1=Proc()
    { p = new ASTSelect(c2.image, c1.image, p1);
      p1.setanc(p);
      }
    |
    <CLOSE> c1=<Id>  { p = new ASTClose(c1.image); }
    |
    <CCLOSE> c1=<Id> <SEMIC> p1 = Proc() { p = new ASTCoClose(c1.image,p1);  p1.setanc(p); }
    |
    <UNFOLD> c1=<Id> <SEMIC> p1 = Proc() { p = new ASTUnfold(c1.image,p1); p1.setanc(p);  }
    |
    <CELL> c1=<Id> <LPAR>
    (LOOKAHEAD(2)
    c2=<Id> ( <COL> t=Type())? <DOT> p1=Proc() <RPAR>
    { p = new ASTCell(c1.image,c2.image,t,p1);  p1.setanc(p); }
    |
    pe = Expr()  <RPAR>
    {
    p = new ASTCell(c1.image,ASTType.gensym(),null,pe);  pe.setanc(p);
    }
    )
    |
    <EMPTY> c1=<Id>
    { p = new ASTEmptyCell(c1.image);}
    |
    <DISCARD> c1 =<Id>
    { p = new ASTDiscard(c1.image); }
    |
    <USE> c1 = <Id> <SEMIC> p1 = Proc()
    {
        p = new ASTUse(c1.image, p1);
        p1.setanc(p);
        // p = p1;
    }
    |
    <AFFINE> c1 = <Id> <SEMIC> p1 = Proc()
    { p = new ASTAffine(c1.image, p1);  p1.setanc(p); }
    |
    <RELEASE> c1=<Id>
    { p = new ASTRelease(c1.image); }
    |
    <PUT> c1=<Id> <LPAR>
    (LOOKAHEAD(2)
    c2=<Id> (<COL> t=Type())? <DOT> p1=Proc() <RPAR> <SEMIC> p2=Proc()
    { p = new ASTPut(c1.image,c2.image,t,p1,p2);  p1.setanc(p);  p2.setanc(p);}
    |
    pe = Expr()  <RPAR> <SEMIC> p2=Proc()
    {
       p = new ASTPut(c1.image,ASTType.gensym(),null,pe,p2);  pe.setanc(p);  p2.setanc(p);
    }
    )
    |
    <TAKE> c1=<Id> <LPAR> c2=<Id> (<COL> t=Type())? <RPAR> <SEMIC> p1=Proc()
    { p = new ASTTake(c1.image,c2.image,t,p1);  p1.setanc(p); }
    |
    <LOCK> c1=<Id> <SEMIC> p1=Proc()
    { p = new ASTLock(c1.image,p1);  p1.setanc(p);}
    |
    <UNLK> c1=<Id> <SEMIC> p1=Proc()
    { p = new ASTUnlock(c1.image,p1);  p1.setanc(p); }
    |
    <WRT> c1=<Id> <LPAR> c2=<Id> <COL> t=Type() <DOT> p1=Proc() <RPAR> <SEMIC> p2=Proc()
    { p = new ASTWrt(c1.image,c2.image,t,p1,p2);  p1.setanc(p);  p2.setanc(p);}
    |
    <RD> c1=<Id> <LPAR> c2=<Id> <COL> t=Type() <RPAR> <SEMIC> p1=Proc()
    { p = new ASTRead(c1.image,c2.image,t,p1);  p1.setanc(p);}
    |
    <FWD> c1=<Id> c2=<Id> { p = new ASTFwd(c1.image, c2.image); }
    |
    <FWDB> c1=<Id> c2=<Id> { p = new ASTFwdB(c1.image, c2.image); }
    |
    <BANG> c1=<Id> <LPAR> c2=<Id> (<COL> t=Type())? <RPAR> <SEMIC> p1=Proc()
    { p = new ASTBang(c1.image,c2.image,t,p1);   p1.setanc(p); }
    |
    <CALL> c1=<Id> <LPAR> c2=<Id> (<COL> t=Type())? <RPAR> <SEMIC> p1=Proc()
    { p = new ASTCall(c1.image,c2.image,t,p1);  p1.setanc(p); }
    |
    <WHY> c1=<Id>  <SEMIC> p1=Proc()
        { p = new ASTWhy(c1.image,p1);  p1.setanc(p) ;
            //p = p1;
        }
    |
    <FOR> c1=<Id> <COL> t=Type() <TO> c2=<Id>
    	  <LPAR>
    	   c3=<Id>
	<COMMA>
	    c4=<Id>
	<SEMIC>
    	   c5=<Id>
	<RPAR>
	<LBRA>
    	  p1=Proc()
	<RBRA>
        { p = ASTNode.ForMacroFactory(c1.image,t,c2.image,c3.image,c4.image,c5.image, p1);
	   p1.setanc(p);  // provisional
	}
    )
    { return p;}
}


void ProcCase(ASTCase lt) throws Exception:
{Token s; ASTNode t;}
{
<VBAR> s=<lab> <COL> t=Proc() {
   lt.addCase(s.image,t);
   t.setanc(lt);
}
}

