
import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.*;
import java.util.logging.*;


class LinSession extends Session implements Cell, Channel {

    private Logger logger;

    private final Lock lock;

    private final Condition condition;

    private boolean message2take, putLock, takeLock;

    private Object message;

    private LinSession fwdSession;

    private boolean cell;

    private int usages;

    private boolean lockCell;

    private  String ch;

    private String chi;

    private  ASTNode rhs;

    private  Env<EnvEntry> ep;

    private  Env<LinSession> ed;

    private Env<Server> eg;

    private static int n = 0;

    private boolean linear;


    LinSession() {
        message2take = false;
        putLock = false;
        takeLock = false;
        cell = false;
        fwdSession = null;
        usages = 1;
        lock = new ReentrantLock();
        condition = lock.newCondition();
    }

    LinSession(String _id) {
        this();
        id = _id + "$" + n++;

	// 	System.out.println("\nSESSION created @ "+id);
    }

    public String getId(){
        return id;
    }

    public void send(Object _message) throws Exception {
        lock.lock();
        while (putLock) {
            condition.await();
        }
        putLock = true;
        message = _message;
        message2take = true;
        condition.signalAll();
        while (message2take && (fwdSession == null)) {
            condition.await();  // !message2take || fwdSession != null
        }
	// !message2take || fwdSession != null
        if (!message2take) {
            putLock = false;
            condition.signalAll();
            lock.unlock();
            return;
        } else {
            fwdSession.send(_message);
            message2take = false;
            putLock = false;
            condition.signalAll();
            lock.unlock();
        }
    }


    public Object receive() throws Exception{
        lock.lock();
	// System.out.println("got lock receive @"+id);
        while(takeLock) { condition.await();}
 	// System.out.println("takelock got lock receive @"+id);
        takeLock = true;
        while(!message2take && (fwdSession == null)) { condition.await(); }
 	// System.out.println("message2take  @"+id);
        if(message2take){
            Object _message = message;
            message2take = false;
            takeLock = false;
            condition.signalAll();
            lock.unlock();
            return _message;
        }else {
            Object _message = fwdSession.receive();
            takeLock = false;
            condition.signalAll();
            lock.unlock();
            return _message;
        }
    }

    public LinSession setFwdSession(LinSession fwdSession){
        lock.lock();
        if(this.fwdSession == null) {
            fwdSession.incUsages(usages - 1);
            this.fwdSession = fwdSession;
            condition.signalAll();
            lock.unlock();
	    return fwdSession;
        }else {
            // this.fwdSession.setFwdSession(fwdSession); // changed
            this.fwdSession = this.fwdSession.setFwdSession(fwdSession); // changed
            lock.unlock();
	    return fwdSession;
        }
        //it was y <-> x and got y <-> z. so x <-> z
    }

    public void setCell(String _chi, ASTNode _rhs, Env<EnvEntry> _ep, Env<LinSession> _ed,
                        Env<Server> _eg, Logger _logger, boolean _linear){
        lock.lock();
        if(fwdSession == null) {
            chi = _chi;
            rhs = _rhs;
            ep = _ep;
            ed = _ed;
            eg = _eg;
            logger = _logger;
            cell = true;
            lockCell = false;
            condition.signalAll();
            lock.unlock();
            linear = _linear;
        }else {
            lock.unlock();
            fwdSession.setCell(_chi, _rhs, _ep, _ed, _eg, _logger, _linear);
        }
    }

    public void setEmptyCell(Logger _logger){
        lock.lock();
        if(fwdSession == null){
            logger = _logger;
            cell = true;
            linear = true;
            lockCell = true;
            condition.signalAll();
            lock.unlock();
        }else {
            lock.unlock();
            fwdSession.setEmptyCell(_logger);
        }

    }


    public LinSession take(String _chi) throws Exception {
        //System.out.println("LinSession: start take " + id);
        lock.lock();
      //  System.out.println("LinSession: Lock the object. waiting for cell or fwd " + id);
        while (!cell && fwdSession == null) {
            condition.await();
        }
	// lock.unlock();
        if(cell){
	    //  lock.lock();
          //  System.out.println("Take: A cell is here. Waiting for the lock. " + id);
            while(lockCell){
                condition.await();
            }
         //   System.out.println("Take: Got the lock. " + id);
            lockCell = true;
            LinSession session = new LinSession(_chi);
            ASTNode rhs_r = rhs;
            Env<EnvEntry> ep_r = ep;
            Env<LinSession> ed_r = ed;
            String chi_r = chi;
            Env<Server> eg_r = eg;
            CLLSj.threadPool.submit(
                    new Runnable() {
                        public void run() {
                            try {
                                rhs_r.runproc(ep_r, ed_r.assoc(chi_r, session), eg_r, logger);
                            } catch (Exception e) {
                                e.printStackTrace(System.out);
                            }
                        }
                    });
            lock.unlock();
            //System.out.println("READ finish" + id);
            return session;
        }
        else {
            //System.out.println("Fwd the read operation on " + id + " to " + fwdSession.getId());
         //   System.out.println("Take: fwd to other session. " + id);
            lock.unlock();
           return fwdSession.take(_chi);
        }
    }

    public void put(String _chi, ASTNode _rhs, Env<EnvEntry> _ep, Env<LinSession> _ed, Env<Server> _eg) throws Exception{
        lock.lock();
	// Thread.sleep(3);
        while (!cell && fwdSession == null) {
            condition.await();
        }
        //lock.unlock();
        if(cell){
	    // lock.lock();
            chi = _chi;
            rhs = _rhs;
            ep = _ep;
            ed = _ed;
            eg = _eg;
            lockCell = false;
            condition.signalAll();
            lock.unlock();
        }
        else {
           lock.unlock();
           fwdSession.put(_chi, _rhs, _ep, _ed, _eg);
        }
    }

    public void free() throws Exception{
        lock.lock();
        while (!cell && fwdSession == null) {
            condition.await();
        }
        if(cell){
            usages = usages - 1;
            if(usages == 0 && linear){ //When all usages are released and the cell is linear, we need to discard its body
                LinSession session = new LinSession(chi);
                Env<LinSession> edCut = ed.assoc(chi,session);

                CLLSj.threadPool.submit( new Runnable(){
                    public void run(){ try {
                        rhs.runproc(ep, edCut, eg, logger);
                    } catch (Exception e) {e.printStackTrace(System.out);} }
                });

                CLLSj.threadPool.submit( new Runnable(){
                    public void run(){ try {
                        (new ASTDiscard(chi)).runproc(ep, edCut, eg, logger);
                    } catch (Exception e) {e.printStackTrace(System.out);} }
                });
            }
            lock.unlock();
            return;
        }
        else {
            lock.unlock();
            fwdSession.free();
        }
    }

    public LinSession read(String _chi) throws Exception {
        //System.out.println("LinSession: reading cell " + id);
        lock.lock();
        while (!cell && fwdSession == null) {
            condition.await();
        }
        // lock.unlock();
        if(cell){
            //  lock.lock();
            //System.out.println("Read: A cell is here. " + id);
            LinSession session = new LinSession(_chi);
            ASTNode rhs_r = rhs;
            Env<EnvEntry> ep_r = ep;
            Env<Server> eg_r = eg;
            String chi_r = chi;
            CLLSj.threadPool.submit(
                    new Runnable() {
                        public void run() {
                            try {
                                rhs_r.runproc(ep_r, new Env<LinSession>(chi_r, session, null), eg_r, logger);
                            } catch (Exception e) {
                                e.printStackTrace(System.out);
                            }
                        }
                    });
            lock.unlock();
            //System.out.println("READ finish" + id);
            return session;
        }
        else {
            //System.out.println("Fwd the read operation on " + id + " to " + fwdSession.getId());
            lock.unlock();
            return fwdSession.read(_chi);
        }
    }

    public void write(String _chi, ASTNode _rhs, Env<EnvEntry> _ep, Env<Server> _eg) throws Exception{
        lock.lock();
        while (!cell && fwdSession == null) {
            condition.await();
        }
        //lock.unlock();
        if(cell){
            // lock.lock();
            chi = _chi;
            rhs = _rhs;
            ep = _ep;
            eg = _eg;
            lock.unlock();
        }
        else {
            lock.unlock();
            fwdSession.write(_chi, _rhs, _ep, _eg);
        }
    }

    public void lock() throws Exception{
        //System.out.println("Trying to lock cell " + id);
        lock.lock();
        while (!cell && fwdSession == null) {
            condition.await();
        }
        //  lock.unlock();

        if(cell){
            //   lock.lock();
            while(lockCell){
                condition.await();
            }
            lockCell = true;
            lock.unlock();
        }
        else {
            //System.out.println("Fwd the lock on " + id + " to " + fwdSession.getId());
            lock.unlock();
            fwdSession.lock();
        }
    }

    public void unlock() throws Exception{
        lock.lock();
        while (!cell && fwdSession == null) {
            condition.await();
        }
        // lock.unlock();
        if(cell){
            //  lock.lock();
            lockCell = false;
            condition.signalAll();
            lock.unlock();
        }
        else {
            lock.unlock();
            fwdSession.unlock();
        }
    }


    public void incUsages(int n){
        lock.lock();
        if(fwdSession != null) {
            lock.unlock();
            fwdSession.incUsages(n);
        }
        else {
            usages = usages + n;
            lock.unlock();
        }
    }
}

